clear all; close all;
% Script generating Phase values from an STFT Magnitude.
% Depending on the operating mode, either the Griffin and Lim Algorithm or 
% Phase Gradient Heap Integration or both are applied. 
%
% PARAMETERS:
%
%	operatingMode: 1 ... GLA initialized with zero
%			       2 ... GLA initialized with the phase from the channel with max.  
%			             stft-magnitude - multichannel input signals only
%			       3 ... GLA performed after PGH
%			       4 ... PGH only
%			       5 ... run 1) through 3) for nice convergence comparison
%
%	saveOutput ... save reconstructed signal after finding a phase
%	Nblk ... blocksize in Samples
%	Nfft ... DFT length in Samples
%	overlap ... overlap factor of neighbouring blocks
%	win_type ... window type, 'Hann' or 'Gauss'
%	enfoceHermitian ... enforce hermitian symmetry during PGH
%	tol ... order of magnitude of the tolerance from the maximum value of
%           the stftMagnitude. Bins with a magnitude below exp(tolerance)
%           times the maximum value are assigned a random Phase during PGH.
%
%	alpha ... STFT magnitude averaging mode, see average_magnitudes.m
%
% DEPENDENCIES:
%
%	GLA.m 
%	PGH.m 
%	average_magnitudes.m 
% 	signal_blocking.m 
%	signal_overlapp_add.m
%
% Authors: Ruediger Fasching, Johanna Kerber, Korbinian Wegler
% Date: 2018

% -------------------%
% --- Parameters --- %
% -------------------%
operatingMode = 5; % 1) GLA initialized with zero
				   % 2) GLA initialized with the phase from the channel with max.  
				   %    stft-magnitude - useful for multichannel input signals only
				   % 3) GLA performed after PGH
				   % 4) PGH only
				   % 5) run 1) through 3) for nice convergence comparison
saveOutput = 0;

Nblk = 4096; % Samples | M
Nfft = 2*Nblk; % Samples
overlap = 4;
win_type = 'Gauss';

maxIts = 100; % maximum number of GLA Iterations
satConv = inf; % dB, satisfying Convergence for GLA

enforceHermitian = 0;
tol = -5; % exponent (power of ten)

alpha = 2; % STFT magnitude averaging mode

% --------------------%
% --- Read Signal --- %
% --------------------%
[filename, filepath] = uigetfile('.wav','select .wav File');
file = strcat(filepath,filename); % Filename und Pfad verbinden

[s, fs] = audioread(file);

% Remove CH 63 & 64 from our recordings (Trackingdata and Reference mic)
% s = s(:,1:end-2);

hop = Nblk*1/overlap; % Samples, hopsize in time | a
% fill up with zeros where window function fades in and out
s = [zeros(Nblk+3*hop,size(s,2)); s; zeros(Nblk+3*hop,size(s,2))]; 

% -------------------%
% --- Processing --- %
% -------------------%
% -- Signal Blocking & Preprocessing -- %
S = signal_blocking(s,hop,Nblk);

switch win_type
    case 'Hann'
        w = cos(pi*(-Nblk/2+1:Nblk/2)'/Nblk).^2;
    case 'Gauss'
        w = gausswin(Nblk+1);
        w = w(2:end);
end


W = repmat(w,[1 size(S,2),size(S,3)]);

% sum of windows for correction term
% ws=signal_overlapp_add(W,hop)/2;
ws = overlap/2; % use simpler method b.c. of line 75

Sw = W.*S;

% - Zero-Padding - %
Sw = [Sw;zeros(Nfft-Nblk,size(Sw,2),size(Sw,3))];
Sw0 = Sw;

% - shift the center of the window to zero phase - %
Sw = circshift(Sw0,-Nfft/2+1);

% -- STFT -- %
Swf = fft(Sw,Nfft);
Phi0 = angle(Swf);

% -- Averaging STFT Magnitude for multichannel signals -- %
if size(s,2) >= 2
    SwfAllCH = Swf;
	[Swf, indexMaxMean] = average_magnitudes(abs(Swf), alpha);
	Phi0 = 0;
else
    indexMaxMean = 1;
end

switch operatingMode
    case 1
    	numGLAs = 1;
    case 2
    	numGLAs = 2;
        iterations = 100;
    case 3
    	numGLAs = 3;
   	case 4
   		numGLAs = 0;
        disp('PGH:')
        tic
		Phi = PGH(abs(Swf), tol, Nblk, hop, Nfft, win_type);
        toc
   	case 5
   		numGLAs = 1:3;
end

% --  GLA - algorithm -- %
% Initialize convergence plot
figure
hold on
grid on
title('Convergence')
ylabel('Normalized abs.-value difference in dB')
xlabel('number  of iterations')

for k = numGLAs
    tic
    switch k
		case 1
	    	Phi = 0;
	    	legendString = 'init. Phase zero';
	    case 2
	    	Phi = angle(SwfAllCH(:,:,indexMaxMean));
	    	 legendString = 'init. Phase max. mean Channel';
	   	case 3
			Phi = PGH(abs(Swf), tol, Nblk, hop, Nfft, win_type);
			legendString = 'init. Phase from PGH';

    end
    disp(legendString);
    
    [Phi, convergence] = GLA( ... 
                         abs(Swf).*exp(1i*Phi), ... % Signal to use GLA on (STFT magnitude and starting Phase)
                         hop, ... 					% Hop size for block processing
                         Nblk, ... 					% block length
                         Nblk/2, ... 				% circular shift length
                         w, ... 					% window function to use
                         maxIts, ...        		% max number of iterations
                         satConv, ...               % break on dB conversion
                         ws ... 					% window correction term 
                     );
	toc;

	% Plot Convergence
    plot(db(convergence),'color',rand(1,3),'DisplayName',legendString)            
end
% enable legend on convergence plot
legend('show')

% -- Resynthesis -- %
% enforce Hermitian symmetry:
if enforceHermitian == 1
    Phi = [Phi; zeros(Nfft-size(Phi,1),size(Phi,2))];
    Phi(Nfft/2+2:Nfft,:)=-flipud(Phi(2:Nfft/2,:));
end
 
Swf = abs(Swf).*exp(1i*Phi);
Sw = real(ifft(Swf));

% shift back from zero phase to original
Sw = circshift(Sw,Nfft/2-1);

% remove zero padding
Sw = Sw(1:Nblk,:,:);

% Overlapp and Add and account for the gain of the sum of the windows
sr = signal_overlapp_add(Sw,hop)./ws;

if saveOutput == 1
	[filename, filepath] = uiputfile('.wav','select .wav File');
	file = strcat(filepath,filename); % Filename und Pfad verbinden
	audiowrite(file,0.8*sr,fs);
end

% --------------%
% --- Plots --- %
% --------------%

if size(s,2) <= 2
    figure
    plot(s)
    hold on
    plot(sr,'r--')
    title('Signal before and after Phase (re)construction')

    figure
    subplot(211)
    plot(s-sr(1:length(s)))
    legend('Difference')
    grid on
    subplot(212)
    plot(s+sr(1:length(s)))
    legend('Sum')
    grid on
end