function  phaseEstimate = PGH(stftMagnitude, tolerance, Nblk, hop, Nfft, win_type)
% Function implementing Phase Gradient Heap Integration to generate a Phase from
% an STFT magnitude.
%
% Based on:
%   [1] Prusa et al. "A Noniterative Method for Reconstruction of Phase from 
%       STFT Magnitude", 
%       IEEE/ACM Transactions on Audio, Speech, and Language Processing        
%       Volume 25 Issue 5 Pages 1154-1164,
%       2017
%
% IN: 
%   stftMagnitude ... stft magnitude in a matrix format. Rows contain samples
%                     in time, columns samples in frequency.
%   tolerance ... order of magnitude of the tolerance from the maximum value of
%                 the stftMagnitude. Bins with a magnitude below exp(tolerance)
%                 times the maximum value are assigned a random Phase.
%   Nblk ... blocksize in Samples
%	hop ... Samples, hopsize in time
%   Nfft ... DFT length in Samples
%	win_type ...window type, 'Hann' or 'Gauss'
%
% OUT:
%   phaseEstimate ... estimated phase values for the given magnitudes.
%
% Authors: Ruediger Fasching, Johanna Kerber, Korbinian Wegler
% Date: 2018

% initialize variables
  numberOfBlocks = size(stftMagnitude,2);  % | N
  freqVec = (0:min(Nfft-1,size(stftMagnitude,1)-1))'; % Vector of frequency steps | m
  
  lambda = hop*Nfft / Nblk; % time frequenzy ratio of window, as in [1], section IV, A  
  switch win_type
      case 'Hann'
          lambda = 2.413*lambda;
      case 'Gauss'
          lambda = 2*lambda;
      otherwise
          error('unknown win_type');
  end
  
  % initialize phase estimates with random values from [0,2*pi]
  phaseEstimate = rand(size(stftMagnitude))*2*pi;
  
  % initialize the heap empty
  heap = [];

  % calculate numerical differentiation of magnitude in time and frequency 
  [timeDiff,freqDiff] = gradient(log(stftMagnitude+exp(-10)));

  % calculate the scaled stft phase gradient (approximation)
  % consisting of the numerical differentiations of the stft magnitude
  % in time and frequency. See (29) in [1]
  frequencyPhaseGradient =  -(lambda/hop * (Nblk/Nfft)).*timeDiff ;
  timePhaseGradient     = (hop/(lambda)*(Nfft/Nblk)).*freqDiff + ...
                          repmat(2*pi*hop*freqVec/Nfft, 1,numberOfBlocks) ;

  % ignore bins with a magnitude below exp(tolerance) times the maximum magnitude
  stftMagnitude(stftMagnitude < exp(tolerance)*max(max(stftMagnitude))) = 0;

  % the following implements Algorithm 1 from [1]
  while any(stftMagnitude(:) ~= 0 ) 
    if isempty(heap)
        [val,idx] = max(stftMagnitude(:));
        [m,n] = ind2sub(size(stftMagnitude),idx);
        heap = [val,m,n];
        
        phaseEstimate(m,n)= 0;
        stftMagnitude(m, n) = 0;
    end
    
    while (~isempty(heap))
      [~,idxmax] = max(heap(:,1));
      m = heap(idxmax,2);
      n = heap(idxmax,3);
      heap(idxmax,:) = [];

      if (m+1 < size(stftMagnitude,1) && stftMagnitude(m+1,n) ~= 0)
        phaseEstimate(m+1, n) = phaseEstimate(m,n) + ... 
                              0.5*  (frequencyPhaseGradient(m,n) + ...
                                frequencyPhaseGradient(m+1, n));
        heap = [heap; stftMagnitude(m+1,n),m+1,n];
        stftMagnitude(m+1,n) = 0;
      end

      if ((m-1 >0)  && (stftMagnitude(m-1,n)~= 0 ))
        phaseEstimate(m-1, n) = phaseEstimate(m,n) - ...
                             0.5*  (frequencyPhaseGradient(m,n) + ...
                                frequencyPhaseGradient(m-1, n));
        heap = [heap; stftMagnitude(m-1,n),m-1,n];
        stftMagnitude(m-1,n) = 0;
      end

      if ((n+1 < size(stftMagnitude,2))  && (stftMagnitude(m,n+1)~= 0) )
        phaseEstimate(m, n+1) = phaseEstimate(m,n) + ...
                        0.5*   (timePhaseGradient(m,n) + ...
                                timePhaseGradient(m, n+1));
        heap = [heap; stftMagnitude(m,n+1),m,n+1];
        stftMagnitude(m,n+1) = 0;
      end

      if ((n-1 >0)  && (stftMagnitude(m,n-1)~= 0) )
        phaseEstimate(m, n-1) = phaseEstimate(m,n) - ...
                      0.5*  (timePhaseGradient(m,n) + ...
                                timePhaseGradient(m, n-1));
        heap = [heap; stftMagnitude(m,n-1),m,n-1];
        stftMagnitude(m,n-1) = 0;
      end        
    end    
  end
end

% EOF